#include "../revengmc.hpp"
#include <cmath>
#include <iostream>
#include <pthread.h>

#define MAX_48 0xFFFFFFFFFFFF
#define MAGIC_NUM 25214903917

#ifndef THREADS
#define THREADS 32
#endif

#define BATCH_SIZE ((size_t)1 << 20)

struct range_t {
	size_t start;
	size_t end;
};

struct impl_t {
	pthread_mutex_t print_lock;

	pthread_t threads[THREADS];
	range_t args[THREADS];

	std::vector<condition_t> conditions;
	std::vector<uint64_t> results;
};

static impl_t *impl;

static uint64_t enchant(uint64_t seed) {
	return (seed ^ MAGIC_NUM) & MAX_48;
}

static uint32_t next_32(size_t &seed, uint32_t max) {
	uint64_t nextseed = (seed * MAGIC_NUM + 11) & MAX_48;
	seed = nextseed;
	return (nextseed >> 17) % max;
}

static bool check_condition(const condition_t &cond, uint64_t seed) {
	uint64_t rand = enchant(cond.part_seed + seed);

	int32_t res_x = cond.res_x;
	int32_t res_y = cond.res_y;

	if (!cond.linear_sep) {
		res_x += (next_32(rand, cond.diff) + next_32(rand, cond.diff)) / 2;
		res_y += (next_32(rand, cond.diff) + next_32(rand, cond.diff)) / 2;
	}
	else {
		res_x += next_32(rand, cond.diff);
		res_y += next_32(rand, cond.diff);
	}

	pthread_mutex_lock(&impl->print_lock);
	// std::cout << res_x << "; " << cond.x << "; " << res_y << "; " << cond.y << std::endl;
	pthread_mutex_unlock(&impl->print_lock);
	return res_x == cond.x && res_y == cond.y;
}

static bool check_seed(uint64_t seed) {
	for (auto it = impl->conditions.begin(); it != impl->conditions.end(); it++) {
		if (!check_condition(*it, seed)) return false;
	}

	return true;
}

void result(uint64_t seed) {
	pthread_mutex_lock(&impl->print_lock);
	impl->results.push_back(seed);
	pthread_mutex_unlock(&impl->print_lock);
}

void check_range(range_t *pargs) {
	auto &args = *pargs;

	for (size_t i = args.start; i < args.end; i++) {
		if (check_seed(i)) result(i);
	}
}

void engine::init(const std::vector<condition_t> &tasks) {
	if (impl != nullptr) {
		std::cerr << "Backend already initialized";
		std::exit(1);
	}

	impl = new impl_t();

	pthread_mutex_init(&impl->print_lock, nullptr);

	impl->conditions = tasks;
}

size_t engine::batch_size() {
	return THREADS * BATCH_SIZE;
}

const std::vector<uint64_t> &engine::run(size_t start, size_t end) {
	pthread_t threads[THREADS];
	range_t args[THREADS];

	size_t n = end - start;
	size_t group_size = n / THREADS;

	for (size_t thread_i = 0; thread_i < THREADS; thread_i++) {
		args[thread_i].start = group_size * thread_i + start;
		args[thread_i].end = group_size * thread_i + start + group_size;
		pthread_create(&threads[thread_i], NULL, (void *(*)(void *))check_range, &args[thread_i]);
	}

	for (size_t i = 0; i < THREADS; i++) {
		pthread_join(threads[i], NULL);
	}

	if (group_size * THREADS < n) {
		auto arg = range_t { .start = group_size * THREADS + start, .end = end };
		check_range(&arg);
	}

	return impl->results;
}

void engine::free() {
	if (impl == nullptr) {
		std::cerr << "Backend already freed";
		std::exit(1);
	}

	pthread_mutex_destroy(&impl->print_lock);
	delete impl;
}
